import pytest
import numpy as np
from   matplotlib.text import Text
from   cachai.gadgets import PolarText

@pytest.fixture
def sample_polartext():
    return PolarText(center=(0, 0), radius=1, angle=45, text="Test", pad=0.1)

class TestPolarText:
    def test_inheritance(self,sample_polartext):
        assert isinstance(sample_polartext, Text),\
                'PolarText do not inherit from matplotlib.text.Text'

    def test_cartesian_coords(self,sample_polartext):
        expected_x = (1 + 0.1) * np.cos(np.pi/4)
        expected_y = (1 + 0.1) * np.sin(np.pi/4)
        actual_x, actual_y = sample_polartext.get_position()
        assert np.isclose(actual_x, expected_x)
        assert np.isclose(actual_y, expected_y)

    def test_properties_set(self,sample_polartext):
        """Verifica que las propiedades se establecen correctamente"""
        test_pt = PolarText(sample_polartext.center, sample_polartext._radius,
                            sample_polartext._angle,text='Test text')
        test_pt.set_polar_position(radius=2.5,angle_deg=25,center=(3,4))
        assert np.array_equal(test_pt.center, np.array([3, 4]))
        assert test_pt._radius == 2.5
        assert np.isclose(test_pt._angle, np.deg2rad(25))
        test_pt.set_pad(0.5)
        assert test_pt._pad == 0.5
        assert test_pt.get_text() == 'Test text'

    def test_add_to_axes(self,figure_with_axes,sample_polartext):
        fig, ax = figure_with_axes
        ax.add_artist(sample_polartext)
        assert sample_polartext in ax.texts,\
                'PolarText is not in the artists list'

    def test_rendering(self,figure_with_axes):
        fig, ax = figure_with_axes
        try:
            texts = [PolarText((0, 0), 1, 0, '0°'),
                     PolarText((0, 0), 1, 90, '90°'),
                     PolarText((0, 0), 1, 180, '180°', pad=0.2)]
            for text in texts:
                ax.add_artist(text)
            fig.canvas.draw()
        except Exception as e:
            pytest.fail(f'PolarText rendering went wrong ({e})')

if __name__ == '__main__':
    pytest.main(['-v', __file__])


# Basic imports
import numpy as np
from   matplotlib import pyplot as plt
# Cachai imports
from   cachai._core.chord import ChordDiagram
from   cachai.gadgets import PolarText
from   cachai.utilities import validate_kwargs

def chord(
        corr_matrix,names=None,colors=None,*,ax=None,radius=1,position=(0,0),optimize=True,
        filter=True,bezier_n=30,show_diag=False,threshold=0.1,node_linewidth=10,node_gap=0.1,
        node_labelpad=0.2,blend=True,blend_resolution=200,chord_linewidth=1,chord_alpha=0.7,
        off_alpha=0.1,positive_hatch=None,negative_hatch='---',fontsize=15,font=None,
        min_dist=np.deg2rad(15),scale='linear',max_rho=0.4,max_rho_radius=0.7,show_axis=False,
        legend=False,positive_label=None,negative_label=None,rasterized=False,**kwargs,
    ):
    """
    A Chord Diagram from a correlation matrix, with customizable threshold, style, and colors.
    
    In science, high-dimensional data are common, the choice of visualization tools directly affects
    both interpretation and communication. Chord diagrams are particularly valuable for illustrating
    weighted, non-directional connections --- such as (anti-)correlations --- between variables,
    treating parameters as nodes and their correlations as links.

        
    Parameters
        corr_matrix : :class:`numpy.ndarray` or :class:`pandas.DataFrame`
            Correlation matrix for the chord diagram. This matrix has to be 2-dimensional, not
            empty, symmetric, and filled just with ``int`` or ``float`` values.
        names / n : :class:`list`, optional
            Names for each node (default: 'Ni' for the i-th node)
        colors / c : :class:`list`, optional
            Custom colors for nodes (default: seaborn hls palette)
        ax : :class:`matplotlib.axes.Axes`, optional
            Axes to plot on (default: current pyplot axis)
    
    Returns
        :class:`ChordDiagram`:viewsource:`cachai._core.chord.ChordDiagram`
            An instance of the ChordDiagram class

    Other Parameters
        radius / r : :class:`float`
            Radius of the diagram (default: 1.0)
        position / p : :class:`tuple`
            Position of the center of the diagram (default: (0,0))
        optimize : :class:`bool`
            Whether to optimize node order (default: True)
        filter : :class:`bool`
            Whether to remove nodes with no correlation (default: True)
        bezier_n : :class:`int`
            Bezier curve resolution (default: 30)
        show_diag : :class:`bool`
            Show self-connections (default: False)
        threshold / th : :class:`float`
            Minimum correlation threshold to display (default: 0.1)
        node_linewidth / nlw : :class:`float`
            Line width for nodes (default: 10)
        node_gap / ngap : :class:`float`
            Gap between nodes (0-1) (default: 0.1)
        node_labelpad / npad : :class:`float`
            Label position adjustment (default: 0.2)
        blend : :class:`bool`
            Whether to blend chord colors (default: True)
        blend_resolution : :class:`int`
            Color blend resolution (default: 200)
        chord_linewidth / clw : :class:`float`
            Line width for chords (default: 1)
        chord_alpha / calpha : :class:`float`
            Alpha of the facecolor for chords (default: 0.7)
        off_alpha : :class:`float`
            Alpha for non-highlighted chords (default: 0.1)
        positive_hatch : :class:`str`
            Hatch for positive correlated chords (default: None)
        negative_hatch : :class:`str`
            Hatch for negative correlated chords (default: '---')
        fontsize : :class:`int`
            Label font size (default: 15)
        font : :class:`dict` or :class:`str`
            Label font parameters (default: None)
        min_dist : :class:`float`
            Minimum angle distance from which apply radius rule (default: 15 [degrees])
        scale : :class:`str`
            Scale use to set chord's thickness, wheter "linear" or "log" (default: "linear")
        max_rho : :class:`float`
            Maximum chord's thickness (default: 0.4) 
        max_rho_radius : :class:`float`
            Maximum normalized radius of the chords relative to center (default: 0.7)
        show_axis : :class:`bool`
            Whether to show the axis (default: False)
        legend : :class:`bool`
            Adds default positive and negative labels in the legend (default: False)
        positive_label : :class:`str`
            Adds positive label in the legend (default: None)
        negative_label : :class:`str`
            Adds negative label in the legend (default: None)
        rasterized : :class:`bool`
            Whether to force rasterized (bitmap) drawing for vector graphics output (default: False)
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    First, import **cachai** and the necessary libraries.
    
    .. code-block:: python
        :class: mock-block

        import matplotlib.pyplot as plt
        import cachai.chplot as chp


    The only mandatory parameter for ``chord()`` is the correlation matrix. Assuming you already
    have your matrix, this line will show the default plot (in some IDLEs you need to
    add ``plt.show()``):

    .. code-block:: python
        :class: mock-block

        chp.chord(corr_matrix)
        

    Now, if you want to add a legend distinguishing positive and negative correlations, just set
    ``legend=True``. Check the next example:

    .. code-block:: python
        :class: mock-block

        plt.figure(figsize=(6,6))

        chp.chord(corr_matrix,
                  threshold=0.3,
                  negative_hatch='///',
                  legend=True,
                  rasterized=True)

        plt.legend(loc='center',bbox_to_anchor=[0.5,0],ncols=2,fontsize=13,handletextpad=0)
        plt.show()

    To see more examples on how to use ``chplot.chord()`` check out the
    :doc:`examples section <../../examples>`.


    Methods
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. currentmodule:: cachai._core.chord

    .. automethod:: ChordDiagram.highlight_node
        :noindex:
    .. automethod:: ChordDiagram.highlight_chord
        :noindex:
    .. automethod:: ChordDiagram.set_chord_alpha
        :noindex:
    """
    # Process parameters
    params = {
        'corr_matrix'      : corr_matrix,
        'names'            : names,
        'colors'           : colors,
        'ax'               : ax if ax is not None else plt.gca(),
        'radius'           : radius,
        'position'         : position,
        'optimize'         : optimize,
        'filter'           : filter,
        'bezier_n'         : bezier_n,
        'show_diag'        : show_diag,
        'threshold'        : threshold,
        'node_linewidth'   : node_linewidth,
        'node_gap'         : node_gap,
        'node_labelpad'    : node_labelpad,
        'blend'            : blend,
        'blend_resolution' : blend_resolution,
        'chord_linewidth'  : chord_linewidth,
        'chord_alpha'      : chord_alpha,
        'off_alpha'        : off_alpha,
        'positive_hatch'   : positive_hatch,
        'negative_hatch'   : negative_hatch,
        'fontsize'         : fontsize,
        'font'             : font,
        'min_dist'         : min_dist,
        'scale'            : scale,
        'max_rho'          : max_rho,
        'max_rho_radius'   : max_rho_radius,
        'show_axis'        : show_axis,
        'legend'           : legend,
        'positive_label'   : positive_label,
        'negative_label'   : negative_label,
        'rasterized'       : rasterized,
    }
    
    # Alternative kwargs aliases
    aliases = {'names'           : 'n',
               'colors'          : 'c',
               'radius'          : 'r',
               'position'        : 'p',
               'threshold'       : 'th',
               'node_linewidth'  : 'nlw',
               'node_gap'        : 'ngap',
               'node_labelpad'   : 'npad',
               'chord_linewidth' : 'clw',
               'chord_alpha'     : 'calpha'
              }
    
    for key in aliases:
        if aliases[key] in kwargs: params[key] = kwargs.pop(aliases[key])

    # Check for wrong kwargs
    validate_kwargs(kwargs.keys(),params.keys(),aliases)

    return ChordDiagram(**params)

def polartext(radius,angle,text,center=(0,0),pad=0.0,**kwargs):
    """
    Place text at specified polar coordinates relative to a center point.

    Parameters
        radius : :class:`float`
            Radius coordinate from the center.
        theta : :class:`float`
            Angle in degrees (0° at positive x-axis, increasing counterclockwise).
        text : :class:`str`
            Text to diplay.
        center : :class:`tuple`, optional
            Center coordinates (x,y) (default: (0,0)).
        pad : :class:`float`, optional
            Padding value (positive for outward displacement, negative for inward).
    
    Returns
        :class:`PolarText`:viewsource:`cachai.gadgets.PolarText`
            An instance of Polar Text (base: :class:`matplotlib.text.Text`)

    Other Parameters
        ``**kwargs``
            Keyword arguments of :class:`matplotlib.text.Text`.
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: mock-block

        import matplotlib.pyplot as plt
        import cachai.chplot as chp

        fig, ax = plt.subplots()
        ax.plot([0,0],[1,1])
        chp.polartext((0,0),1,45,text='Im polar!',pad=0.5,ha='center',va='center')
        plt.show()

    To see more examples check out the :doc:`examples section <../../examples>`.

    Methods
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. currentmodule:: cachai.gadgets

    .. automethod:: PolarText.set_polar_position
        :noindex:
    .. automethod:: PolarText.set_pad
        :noindex:
    .. automethod:: PolarText.set_radius
        :noindex:
    .. automethod:: PolarText.set_angle
        :noindex:
    .. automethod:: PolarText.set_center
        :noindex:
    
    Other Methods
        Inherited from :class:`matplotlib.text.Text`.
    """
    ax     = plt.gca()
    artist = PolarText(radius,angle,text,center,pad,**kwargs)
    ax.add_artist(artist)
    return artist


import sys
import argparse
from cachai.tests._run_test import run_tests

def main():
    parser = argparse.ArgumentParser(
        description='Run the tests from CACHAI',
        usage='%(prog)s [OPTIONS] [TESTS...]'
    )
    
    # For specific tests
    parser.add_argument(
        'tests',
        nargs='*',
        default=[],
        help="Names of the tests to run (e.g. 'charts', 'utilities'). If not specified, all tests are run."
    )
    
    # Option to show details
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Show detailed output (equivalent to pytest -v)'
    )
    
    # Option to show available tests
    parser.add_argument(
        '-l', '--list',
        action='store_true',
        help='List all available tests without running them'
    )
    
    args = parser.parse_args()
    
    if args.list:
        from cachai.tests._run_test import get_available_tests
        print('Available tests:')
        for test in get_available_tests():
            print(f'  {test}')
        sys.exit(0)
    
    sys.exit(run_tests(*args.tests))

if __name__ == '__main__':
    main()


# Basic imports
import numpy as np
import pandas as pd
import seaborn as sns
import cachai.utilities as chu
import cachai.gadgets as chg
# Matplotlib imports
from   matplotlib import pyplot as plt
from   matplotlib.patches import Arc, Circle, PathPatch
from   matplotlib.path import Path
import matplotlib.colors as mtpl_colors
from   matplotlib.text import Text

class ChordDiagram():
    def __init__(self, corr_matrix, **kwargs):
        """
        Initialize a ChordDiagram instance.
        """
        # Correlation matrix error handling
        self.corr_matrix = corr_matrix
        self._validate_corr_matrix()
        
        # Initialize additional parameters
        self.__dict__.update(kwargs)
        if isinstance(self.corr_matrix, pd.DataFrame):
            if self.names is None: self.names = self.corr_matrix.columns.tolist()
            self.corr_matrix = self.corr_matrix.to_numpy()
        if self.names is None: self.names = [f'N{i+1}' for i in range(len(self.corr_matrix))]
        if self.colors is None: self.colors = sns.hls_palette(len(self.corr_matrix))
        self.nodes = dict()
        self.order = [i for i in range(len(self.corr_matrix))]
        self.global_indexes = []
        if self.font is None: self.font = {'size':self.fontsize}
        
        # Initialize collection lists
        self.node_patches        = []
        self.node_labels         = []
        self.node_labels_params  = []
        self.chord_patches       = [[] for i in range(len(self.corr_matrix))]
        self.chord_blends        = [[] for i in range(len(self.corr_matrix))]
        self.bezier_curves       = [[] for i in range(len(self.corr_matrix))]
        self.__ports_refs        = []
        self.__highlighted_ports = []
        
        # Generate the diagram
        self.__generate_diagram()

    # Util methods
    def _validate_corr_matrix(self):
        """
        Validate that a correlation matrix meets the required specifications:
            - Input is a numpy.ndarray or pandas.DataFrame
            - Matrix is 2-dimensional
            - Matrix is not empty
            - All values are int or float
            - Matrix is symmetric
        """
        temp_corr_matrix = self.corr_matrix
        if not isinstance(temp_corr_matrix, (np.ndarray, pd.DataFrame)):
            raise TypeError('Your correlation matrix must be a numpy.ndarray or pandas.DataFrame')
        # -- This block of code should not be here, but its necessary for the next validations --
        if isinstance(temp_corr_matrix, pd.DataFrame):
            temp_corr_matrix = self.corr_matrix.to_numpy()
        # ---------------------------------------------------------------------------------------
        if temp_corr_matrix.ndim != 2:
            raise ValueError('Your correlation matrix must be a 2-dimensional array')
        if temp_corr_matrix.shape[0] != temp_corr_matrix.shape[1]:
            raise ValueError('Your correlation matrix must be a square matrix.')
        if len(temp_corr_matrix) == 0:
            raise ValueError('Your correlation matrix cannot be empty')
        if not np.issubdtype(temp_corr_matrix.dtype, np.floating):
            raise TypeError('Your correlation matrix must contain float values')
        if not np.allclose(temp_corr_matrix, temp_corr_matrix.T):
            raise ValueError('Your correlation matrix must be symmetric')
        
    def _optimize_nodes(self):
        """Optimize node order using Prim's algorithm."""
        n_nodes = self.corr_matrix.shape[0]
        # We convert the correlations to distances
        # The strongest the correlation, the shorter the distance
        distance_matrix = 1 - np.abs(self.corr_matrix)
        # In order to ignore the diagonal, we fill it with infinity values
        np.fill_diagonal(distance_matrix, np.inf)
        
        # Prim's algorithm
        visited = set()
        order  = []

        start_node = np.argmin(np.sum(distance_matrix, axis=0))
        visited.add(start_node)
        order.append(start_node)

        while len(visited) < n_nodes:
            # Closest non-visited node to any visited node
            min_dist = np.inf
            next_node = -1
            for node in visited:
                for neighbor in range(n_nodes):
                    if (neighbor not in visited) and (distance_matrix[node,neighbor] < min_dist):
                        min_dist = distance_matrix[node, neighbor]
                        next_node = neighbor
            if next_node == -1:
                break  # Just in case somehow we have disconnected nodes
            visited.add(next_node)
            order.append(next_node)

        # Apply new order
        self.order = order
        self.__order_nodes()
    
    def _radius_rule(self, dist):
        """Rule to set the radius of a single chord"""
        if dist <= self.min_dist:
            return self.max_rho_radius
        else:
            return self.max_rho_radius * (1 - (dist - self.min_dist) / (np.pi - self.min_dist))
    
    def _scale_rho(self, rho):
        """Scale rho (link thickness)"""
        if self.scale == 'linear':
            rho_lin = np.abs(rho) * self.max_rho
            return np.clip(rho_lin, 0, 1) # Clip to avoid numerical issues
        elif self.scale == 'log':
            rho_log = (1 - np.log10(10 - 9*np.abs(rho))) * self.max_rho
            return np.clip(rho_log, 0, 1) # Clip to avoid numerical issues
        else:
            raise ValueError(f'Unknown scale type {self.scale}')
    
    # Main generation methods
    def __generate_diagram(self):
        """Generate the complete chord diagram"""
        
        for i,color in enumerate(self.colors):
            if isinstance(color,str): self.colors[i] = mtpl_colors.to_rgb(color)
            
        if self.filter == True: self.__filter_nodes()
        
        if len(self.corr_matrix) == 0:
            raise ValueError(f'No nodes remaining after threshold filtering: '
                f'all correlations were below the threshold = {self.threshold}.')
        else:
            if self.optimize == True: self._optimize_nodes()
            self.__generate_nodes()
            self.__generate_chords()

            # Add patches to axes
            for node_patch, node_label in zip(self.node_patches, self.node_labels_params):
                self.ax.add_patch(node_patch)
                label = chg.PolarText(
                    self.position,
                    node_label['r'],
                    np.rad2deg(node_label['theta']),
                    text=node_label['label'],
                    pad=self.node_labelpad,
                    rotation=node_label['rot'],
                    ha='center', va='center',
                    clip_on=True,
                    rasterized=self.rasterized,
                )
                label.set_font(self.font)
                self.ax.add_artist(label)
                self.node_labels.append(label)
            flat_chord_patches = [p for plist in self.chord_patches for p in plist]
            flat_bezier_curves = [c for clist in self.bezier_curves for c in clist]
            for k,(chord_patch,bezier_curve) in enumerate(zip(flat_chord_patches,flat_bezier_curves)):
                self.ax.add_patch(chord_patch)
                if self.blend:
                    self.__add_chord_blend(chord_patch,bezier_curve,self.global_indexes[k])
            
            self.__adjust_ax()
            self.__generate_legend()
            self.__generate_port_refs()
                
    # Components generation methods
    def __generate_nodes(self):
        """Generate nodes"""
        # Initial variables
        # Minus 1 from each diagonal of A to A
        relevance      = np.sum(np.abs(self.corr_matrix),axis=1) - 1
        relevance_norm = relevance / np.sum(relevance)
        start_angles   = np.cumsum([0] + list(2*np.pi*relevance_norm[:-1]))
        gap_angle      = (2*np.pi/len(self.corr_matrix))*self.node_gap

        # Base patch
        self.ax.add_patch(Circle(self.position,self.radius,
                                 lw=0,
                                 zorder=2,
                                 fc='w',
                                 ec='none',
                                 rasterized=self.rasterized))

        lw = 2*self.node_linewidth

        for node in range(len(self.corr_matrix)):
            node_data = dict()
            theta_i   = start_angles[node]
            theta_f   = theta_i + 2*np.pi*relevance_norm[node]
            # -- Gap correction -----
            theta_i   = theta_i + np.min([gap_angle,theta_f])
            # -----------------------
            theta_m   = (theta_i + theta_f)/2
            theta_arc = chu.angdist(theta_i,theta_f)
            node_data['theta_i']   = theta_i
            node_data['theta_f']   = theta_f
            node_data['theta_m']   = theta_m
            node_data['theta_arc'] = theta_arc

            rhos       = dict() # Correlations
            ports      = dict() # Ports of the node
            states     = dict() # Ports states (1 or -1)
            corr       = np.insert(self.corr_matrix[node],node,1)
            real_corr  = corr.copy()
            # Control of the allowed ports using the correlation factor
            # 1 = Allowed
            # -1 = Forbidden
            node_ports_state = [1 for p in range(len(corr))]
            for p,r in enumerate(corr):
                if np.abs(r) < self.threshold:
                    node_ports_state[p] = -1
                    real_corr[p]        = 0
            if not self.show_diag:
                node_ports_state[node]   = -1
                real_corr[node]          = 0
                node_ports_state[node+1] = -1
                real_corr[node+1]        = 0
            if np.sum(real_corr) == 0: corr_norm = real_corr
            else: corr_norm = np.abs(real_corr)/np.sum(np.abs(real_corr))

            for j,(rho,port_size,port_state) in enumerate(zip(corr,corr_norm,node_ports_state)):
                if   j == node     : port_id = node
                elif j == (node+1) : port_id = f'{node}*'
                else:
                    if   j < node: port_id = j
                    elif j > node: port_id = j-1
                port_i = theta_i + theta_arc*np.sum(corr_norm[:j])
                port_f = port_i + theta_arc*port_size
                if port_state < 0:
                    port_i = 0
                    port_f = 0
                rhos[port_id]   = rho
                ports[port_id]  = {'i':port_i,'f':port_f}
                states[port_id] = port_state

            node_data['rhos']        = rhos
            node_data['ports']       = ports
            node_data['ports_state'] = states
            self.nodes[node]         = node_data

        # Node
        for n in self.nodes:
            node = self.nodes[n]

            # Patch
            self.node_patches.append(
                Arc(self.position,
                width=2*self.radius, 
                height=2*self.radius,
                theta1=np.rad2deg(node['theta_i']), 
                theta2=np.rad2deg(node['theta_f']),
                lw=lw,zorder=1, 
                rasterized=self.rasterized,
                color=self.colors[n])
            )

            # Label
            params = dict()
            params['label'] = self.names[n]
            params['r']     = self.radius
            params['theta'] = node['theta_m']
            params['x']     = params['r'] * np.cos(params['theta']) + self.position[0]
            params['y']     = params['r'] * np.sin(params['theta']) + self.position[1]
            params['rot']   = np.rad2deg(node['theta_m'] - np.sign(params['y']-self.position[1])*np.pi/2)%360
            self.node_labels_params.append(params)
        
    def __generate_chords(self):
        """Generate chords"""
        for n in self.nodes:
            node = self.nodes[n]
            chord_color = self.colors[n]
            chord_edge  = chu.mod_color(self.colors[n],light=0.5)
            if self.blend:
                chord_color = 'none'
                chord_edge = '#3D3D3D'

            # Links
            if self.show_diag:
                points,codes,curve = self.__compute_bezier_curves(
                                         (node['ports'][n]['i'],node['ports'][n]['f']),
                                         (node['ports'][f'{n}*']['i'],node['ports'][f'{n}*']['f']),
                                         self._scale_rho(1)
                                     )

                self.chord_patches[n].append(
                    PathPatch(Path(points, codes),
                              facecolor=chord_color,
                              edgecolor=chord_edge,
                              alpha=self.chord_alpha,
                              hatch=self.positive_hatch,
                              lw=self.chord_linewidth,
                              rasterized=self.rasterized,
                              zorder=4)
                )
                curve['c1'] = self.colors[n]
                curve['c2'] = self.colors[n]
                self.bezier_curves[n].append(curve)
                self.global_indexes.append(n)

            for m in self.nodes:
                if m > n and node['ports_state'][m] > 0:
                    try:
                        target   = self.nodes[m]
                        this_rho = node['rhos'][m]
                        vis_rho  = self._scale_rho(this_rho)
                        hatch    = self.positive_hatch
                        if this_rho < 0: hatch = self.negative_hatch

                        points,codes,curve = self.__compute_bezier_curves(
                                         (node['ports'][m]['i'],node['ports'][m]['f']),
                                         (target['ports'][n]['i'],target['ports'][n]['f']),
                                         vis_rho
                                     )

                        self.chord_patches[n].append(
                            PathPatch(Path(points, codes),
                                      facecolor=chord_color,
                                      edgecolor=chord_edge,
                                      alpha=self.chord_alpha,
                                      hatch=hatch,
                                      lw=self.chord_linewidth,
                                      rasterized=self.rasterized,
                                      zorder=4)
                        )
                        curve['c1'] = self.colors[n]
                        curve['c2'] = self.colors[m]
                        self.bezier_curves[n].append(curve)
                        self.global_indexes.append(n)
                        
                    except Exception as e:
                        print(chu.strcol(rf'ChordError: Problem creating chord from {self.names[n]} to {self.names[m]}.',
                                          c='red'))
                        print(chu.strcol(f'            details: {e}',
                                          c='red'))
    
    def __generate_legend(self):
        """Add dummie labels to show in the legend"""
        if self.legend is True:
            if self.positive_label is None: self.positive_label = 'Positive\ncorrelation'
            if self.negative_label is None: self.negative_label = 'Negative\ncorrelation'
        # Dummies
        if self.positive_label is not None:
            dummy = self.ax.scatter(*self.position,marker='s',s=200,
                                    c='lightgray',ec='k',hatch=self.positive_hatch,
                                    label=self.positive_label,zorder=0,rasterized=True)
            #dummy.set_visible(False)
        if self.negative_label is not None:
            dummy = self.ax.scatter(*self.position,marker='s',s=200,
                                    c='lightgray',ec='k',hatch=self.negative_hatch,
                                    label=self.negative_label,zorder=0,rasterized=True)
            #dummy.set_visible(False)
    
    def __generate_port_refs(self):
        for n in self.nodes:
            self.__ports_refs.append(self.__get_node_ports_references(n))

    # Helper methods
    def __filter_nodes(self):
        """Remove nodes with no correlation (0 chords)"""
        mask = np.all((np.abs(self.corr_matrix) < self.threshold)\
                      | (np.eye(self.corr_matrix.shape[0], dtype=bool)), axis=1)
        indexes = np.where(~mask)[0]
        
        self.corr_matrix = self.corr_matrix[indexes][:, indexes]
        self.names       = [self.names[i] for i in indexes]
        self.colors      = [self.colors[i] for i in indexes]
        
    def __order_nodes(self):
        """Order nodes (matrix), names and colors"""
        self.corr_matrix = self.corr_matrix[np.ix_(self.order, self.order)]
        self.names       = [self.names[i] for i in self.order]
        self.colors      = [self.colors[i] for i in self.order]
    
    def __compute_bezier_curves(self,alpha,beta,rho):
        """Compute bezier curves to modelate a chord"""
        # Polar
        alpha_i, alpha_f = alpha
        alpha_m = np.mean([alpha_f,alpha_i])
        alphas = chu._angspace(alpha_i,alpha_f)
        if len(alphas) == 0: alphas = np.array([alpha_i,alpha_f]) # Case: angdist too short

        beta_i, beta_f = beta
        beta_m = np.mean([beta_f,beta_i])
        betas = chu._angspace(beta_i,beta_f)
        if len(betas) == 0: betas = np.array([beta_i,beta_f]) # Case: angdist too short

        dist = chu.angdist(alpha_m, beta_m)
        r_rho = self._radius_rule(dist) * self.radius
        dist_inex = np.min([chu.angdist(alpha_i, beta_f), chu.angdist(alpha_f, beta_i)])

        # Convex case
        if chu.angdist(alpha_i, beta_f) < chu.angdist(alpha_f, beta_i):
            theta_rho = beta_f + dist_inex / 2
            r_AB = r_rho
            r_BA = r_rho + rho * self.radius
        # Concave case
        elif chu.angdist(alpha_i, beta_f) >= chu.angdist(alpha_f, beta_i):
            theta_rho = alpha_f + dist_inex / 2
            r_AB = r_rho + rho * self.radius
            r_BA = r_rho

        # Cartesian
        points_A = np.column_stack([np.cos(alphas) * self.radius, 
                                   np.sin(alphas) * self.radius])
        points_B = np.column_stack([np.cos(betas) * self.radius, 
                                   np.sin(betas) * self.radius])

        # A to B
        point_AB = [np.array([r_AB * np.cos(theta_rho), 
                             r_AB * np.sin(theta_rho)])]
        control_AB = [2 * point_AB[0] - (points_A[-1] + points_B[0]) / 2]

        # B to A
        point_BA = [np.array([r_BA * np.cos(theta_rho), 
                             r_BA * np.sin(theta_rho)])]
        control_BA = [2 * point_BA[0] - (points_A[0] + points_B[-1]) / 2]

        # Bezier curve in the middle
        mid_bezier = dict()
        r_mid = (r_AB + r_BA) / 2
        point_mid = [np.array([r_mid * np.cos(theta_rho), 
                              r_mid * np.sin(theta_rho)])]
        control_mid = [2 * point_mid[0] - (points_A[-1] + points_B[0]) / 2]
        mid_bezier['P0'] = points_A[-1] + self.position
        mid_bezier['P1'] = control_mid[0] + self.position
        mid_bezier['P2'] = points_B[0] + self.position

        # Points
        points = np.vstack((points_A, control_AB, points_B, control_BA, points_A[0])) \
                 + self.position

        # Codes
        codes = [Path.MOVETO] + \
                [Path.LINETO] * (len(points_A) - 1) + \
                [Path.CURVE3] * 2 + \
                [Path.LINETO] * (len(points_B) - 1) + \
                [Path.CURVE3] * 2

        return points,codes,mid_bezier
    
    def __add_chord_blend(self,patch,curve,n):
        """Add color mapped patches using the initial and final colors"""
        # Pach vertices
        vertices = patch.get_path().vertices
        xmin, ymin = np.min(vertices, axis=0)
        xmax, ymax = np.max(vertices, axis=0)

        # Bézier curve
        P0 = curve['P0']
        P1 = curve['P1']
        P2 = curve['P2']
        bezier = chu.get_bezier_curve(P0,P1,P2,n=self.bezier_n)
        bezier_equidistant = chu.equidistant(bezier)

        # Color map
        c1          = curve['c1'] # Color 1
        c2          = curve['c2'] # Color 2
        chord_cmap  = sns.blend_palette([c1,c1,c2,c2],as_cmap=True)
        cmap_matrix = chu.map_from_curve(bezier_equidistant,xlim=(xmin,xmax),ylim=(ymin,ymax),
                                         resolution=self.blend_resolution)
        self.chord_blends[n].append(
            chu.colormapped_patch(
                patch,
                cmap_matrix,
                ax=self.ax,
                colormap=chord_cmap,
                zorder=2,
                alpha=self.chord_alpha,
                rasterized=self.rasterized)
        )
    
    def __adjust_ax(self):
        """Adjust scale, limits, and visibility of the axis"""
        adjust_x = self.ax.get_autoscalex_on()
        adjust_y = self.ax.get_autoscaley_on()
        if adjust_x:
            self.ax.set_xlim(self.position[0] - self.radius*1.5,self.position[0] + self.radius*1.5)
        if adjust_y:
            self.ax.set_ylim(self.position[1] - self.radius*1.5,self.position[1] + self.radius*1.5)
        if adjust_x and adjust_y: self.ax.set_aspect('equal')
        if self.show_axis == False: self.ax.axis('off')
    
    def __get_node_ports(self,n):
        """Return the occupied ports of the n-th node"""
        this_items = list(self.nodes[n]['ports_state'].items())
        this_items.pop(n+1)
        return [p for p, s in this_items if s > 0 and p > n]

    def __get_node_ports_references(self,n):
        """
        Return the reference of the chords of the n-th node as (n,c), where:

        - n: index of the n-th node in the resulting diagram
        - c: index of the c-th chord in the n-th node

        Always anti-clockwise. When show_diag=True, the self-referencing chord is (n,0).
        """
        node       = self.nodes[n]
        this_ports = self.__get_node_ports(n)
        refs       = []
        for port in node['ports_state'].keys():
            if '*' not in str(port):
                if node['ports_state'][port] > 0 and port > n:
                    refs.append((n,this_ports.index(port)))
                elif node['ports_state'][port] > 0 and port < n:
                    target_ports = self.__get_node_ports(port)
                    refs.append((port,target_ports.index(n)))
        if self.show_diag:
            diag_ref_position = None
            refs_modified = []
            for i, (x, y) in enumerate(refs):
                modified = (x, y + 1)
                refs_modified.append(modified)
                if diag_ref_position is None and x == n: diag_ref_position = i   
            if diag_ref_position is None: diag_ref_position = len(refs_modified)
            refs_modified.insert(diag_ref_position, (n, 0))
            refs = refs_modified
        return refs

    def __update_highlights(self):
        for n in self.nodes:
            for c in range(len(self.chord_patches[n])):
                if (n,c) not in self.__highlighted_ports:
                    self.chord_patches[n][c].set_alpha(self.off_alpha)
                    if self.blend == True: self.chord_blends[n][c].set_alpha(self.off_alpha)

    # Customization methods
    def highlight_node(self,node,chords=None,alpha=None):
        """:meta private:
        Highlights a specific node.
        This affects the node and all its chords. If you want to highlight only some chords in the
        node, you can indicate this with ``chords``.

        Nodes are indexed based on their circular arrangement around the origin (center of the
        Chord Diagram). Index ``0`` corresponds to the first node in the first quadrant, with
        numbering proceeding counterclockwise around the diagram. Chords within each node are also
        indexed counterclockwise, starting from the outermost chord.

        Parameters
            node : :class:`int`
                Index of the node to highlight (starting in 0).
            chords : :class:`list` or :class:`array-like`, optional
                List of chord indices to highlight (default: all chords).
            alpha : :class:`float`, optional
                Transparency level for highlighting.
        """
        if node >= len(self.nodes):
            raise IndexError('Node is out of range. '
                f'This Chord Diagram has only {len(self.nodes)} nodes.')
        if node < 0 :
            raise ValueError('The input node must be positive or zero.')

        if chords is None: chords = [c for c in range(len(self.__ports_refs[node]))]
        for chord in chords:
            self.highlight_chord(node,chord,alpha)

    def highlight_chord(self,node,chord,alpha=None):
        """:meta private:
        Highlights a specific chord connected to a particular node.

        Parameters
            node : :class:`int`
                Index of the node where the chord originates.
            chord : :class:`int`
                Index of the chord to highlight (starting in 0).
            alpha : :class:`float`, optional
                Transparency level for highlighting.
        """
        if node >= len(self.nodes):
            raise IndexError('Node is out of range. '
                f'This Chord Diagram has only {len(self.nodes)} nodes.')
        if node < 0 or chord < 0:
            raise ValueError('The input node and chord must be positive or zero.')

        if alpha is None: alpha = self.chord_alpha
        if alpha <= self.off_alpha: alpha = 0.8

        self.__update_highlights()

        try:
            n,c = self.__ports_refs[node][chord]
            self.chord_patches[n][c].set_alpha(alpha)
            if self.blend == True: self.chord_blends[n][c].set_alpha(alpha)
            if (n,c) not in self.__highlighted_ports: self.__highlighted_ports.append((n,c))
        except IndexError:
            raise IndexError(f'Chord {chord} is out of range. '
                    f'Node {node} has only {len(self.__ports_refs[node])} chords.')

    def set_chord_alpha(self,alpha):
        """:meta private:
        Sets the transparency level for all chords in the diagram.

        Parameters
            alpha : :class:`float`
                Transparency value applied to all chords.
        """
        for n in self.nodes:
            for cp in self.chord_patches[n]: cp.set_alpha(alpha)
            if self.blend == True:
                for cb in self.chord_blends[n]: cb.set_alpha(alpha)

    # Special methods 
    def __str__(self):
        string = ''
        for n in self.nodes:
            string += f'node {n} "{self.names[n]}"\n' + '-'*50
            for key in self.nodes[n]:
                if key == 'ports':
                    string += f'\n{key:<10} :'
                    for p in self.nodes[n][key]:
                        string += f'\n\t\t{p:<10} : {self.nodes[n][key][p]}'
                else:
                    string += f'\n{key:<10} : {self.nodes[n][key]}'
            string += '\n\n\n'
        return string


# Imports
import os
import numpy as np
import colorsys
# Matplotlib imports
from   matplotlib import pyplot as plt
# Scipy imports
from   scipy.spatial.distance import cdist
from   scipy.interpolate import interp1d

def chsave(name="figure",dir_path="images",pdf=True,img_dpi=300,pdf_dpi=200):
    """
    Saves the current matplotlib figure as PNG and optionally PDF with customizable *dpi*
    (dots per inch).

    Parameters
        name : :class:`str`
            The name of the figure. The image is saved as "dir_path/name.png" (default: "figure").
    
    Other Parameters
        dir_path : :class:`str`
            Path to the directory where the image is saved. If it doesn't exist, it will be created
            (default: "images")
        pdf : :class:`bool`
            Whether to also save the image as a PDF.
        img_dpi : :class:`int`
            The *dpi* of the PNG image.
        pdf_dpi : :class:`int`
            The *dpi* of the PDF image.
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: mock-block

        import cachai.utilities as chu

        plt.figure(figsize=(10,10))
        plt.scatter([1,2,3],[1,4,9],color="magenta")

        chu.chsave("my_plot")
    """
    if not os.path.exists(dir_path): os.makedirs(dir_path)
    plt.savefig(os.path.join(dir_path,f'{name}.png'),
                bbox_inches='tight',pad_inches=0.3,dpi=img_dpi)
    if pdf:
        if not os.path.exists(os.path.join(dir_path,'pdf')):
            os.makedirs(os.path.join(dir_path,'pdf'))
        plt.savefig(os.path.join(dir_path,'pdf',f'{name}.pdf'),
                    bbox_inches='tight',pad_inches=0.3,dpi=pdf_dpi)

def angdist(alpha,beta):
    """
    Calculates the minimal angular distance between two angles in radians.

    Parameters
        alpha : :class:`float`
            Angle in radians.
        beta : :class:`float`
            Angle in radians.
    
    Returns
        :class:`float`

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import numpy as np

        A = np.pi/4
        B = 3*np.pi/4
        
        distance = chu.angdist(A,B)

        print("Distance in radians:",distance)
        print("Distance in degrees:",np.rad2deg(distance))
    
    .. code-block:: text
        :class: out-block

        Distance in radians: 1.5707963267948966
        Distance in degrees: 90.0
    """
    diff = np.abs(alpha - beta) % (2 * np.pi)
    return np.min([diff, 2 * np.pi - diff])

def _angspace(alpha,beta,n=200):
    """:meta-private:
    Generate a linear space of angles
    """
    theta = abs(beta-alpha)
    ndots = int(theta*n/(2*np.pi))
    if ndots == 1: ndots = 2
    return np.linspace(alpha,beta,ndots)

def map_from_curve(curve=None,xlim=(-1,1),ylim=(-1,1),resolution=200):
    """
    Generates a map (2D matrix) where each point value is based on its proximity to the nearest
    point along a specified curve.

    Parameters
        curve : :class:`numpy.ndarray`
            1D array of points (x,y) defining the reference curve.
        xlim : :class:`tuple` or :class:`array-like`, optional
            x-axis boundaries of the map.
        ylim : :class:`tuple` or :class:`array-like`, optional
            y-axis boundaries of the map.
        resolution : :class:`int`, optional
            Number of grid points along each axis for the output map.
    
    Returns
        :class:`numpy.ndarray` : 2D array

        or

        :class:`None` : If ``curve = None``

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import numpy as np

        # Constructing a sine curve
        x_array = np.linspace(0,2*np.pi,500)
        y_array = np.sin(x_array)
        curve   = np.array([(x,y) for x,y in zip(x_array,y_array)])

        my_map = chu.map_from_curve(curve,xlim=(0,2*np.pi),ylim=(-1,1),resolution=5)

        print(my_map)
    
    .. code-block:: text
        :class: out-block

        [[-1.         -0.10220441  0.15430862  0.498998    0.84769539]
         [-1.         -0.77955912  0.07815631  0.498998    0.91983968]
         [-1.         -0.498998    0.00200401  0.498998    1.        ]
         [-0.91983968 -0.498998   -0.07815631  0.77955912  1.        ]
         [-0.84769539 -0.498998   -0.15430862  0.10220441  1.        ]]
    """
    if curve is None: return None
    
    # Values from curve
    values = np.linspace(-1,1,len(curve))
    
    # Mesh
    x = np.linspace(*xlim, resolution)
    y = np.linspace(*ylim, resolution)
    grid_x, grid_y = np.meshgrid(x, y, indexing='xy')
    grid_points = np.column_stack((grid_x.ravel(), grid_y.ravel()))

    # Obtain the nearest point in the curve and use that value
    distances             = cdist(grid_points, curve)
    nearest_point_indexes = np.argmin(distances, axis=1)
    map_flat              = values[nearest_point_indexes]
    map_matrix            = map_flat.reshape(resolution, resolution)
    
    return map_matrix

def colormapped_patch(patch,map_matrix,ax=None,colormap="coolwarm",
                      zorder=5,alpha=0.5,rasterized=False):
    """
    Applies a colormap to a patch object using a precomputed map matrix, creating a color-filled
    shape.

    Parameters
        patch : :class:`matplotlib.patches.Patch` and similar
            Matplotlib patch object to be filled with colors.
        map_matrix : :class:`numpy.ndarray`
            2D array containing color values for the mapping.
            
    Returns
        :class:`matplotlib.image.AxesImage`

    Other Parameters
        ax : :class:`matplotlib.axes.Axes`
            Axes object where the patch will be drawn (default: current axes).
        colormap : :class:`str` or :class:`matplotlib.colors.LinearSegmentedColormap`
            Matplotlib colormap to use.
        zorder : :class:`int`
            Rendering order layer for the patch.
        alpha : :class:`float`
            Transparency level of the filled patch.
        rasterized : :class:`bool`
            Whether to rasterize the patch for better performance.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: mock-block

        import cachai.utilities as chu
        import numpy as np
        import matplotlib.pyplot as plt
        from   matplotlib.patches import Circle

        # Create a circular patch
        circle = Circle((0.5, 0.5), 0.3)
        # Generate a simple gradient map
        gradient = np.linspace(-1, 1, 400).reshape(20, 20)

        fig, ax = plt.subplots()
        ax.add_patch(circle)
        chu.colormapped_patch(circle, gradient, ax=ax)
        plt.show()
    """
    if ax is None: ax = plt.gca()
    
    vertices   = patch.get_path().vertices
    xmin, ymin = np.min(vertices, axis=0)
    xmax, ymax = np.max(vertices, axis=0)
    
    img = ax.imshow(
        map_matrix, 
        cmap=colormap, 
        extent=(xmin, xmax, ymin, ymax), 
        origin='lower',
        aspect='auto',
        clip_path=patch,
        clip_on=True,
        zorder=zorder,
        alpha=alpha,
        rasterized=rasterized,
        vmin=-1, vmax=1
    )
    return img

def equidistant(points):
    """
    Resamples points along a curve to make them equidistant while preserving the overall shape.

    Parameters
        points : :class:`numpy.ndarray`
            1D array of points (x,y) defining the reference curve.
    
    Returns
        :class:`numpy.ndarray` : 2D array
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import numpy as np

        # Create a non-uniform curve
        original_points = np.array([[0, 0], [1, 2], [3, 1], [4, 4]])

        # Make points equidistant
        equal_points = chu.equidistant(original_points)

        print(equal_points)
        
    .. code-block:: python
        :class: out-block

        [[0.         0.        ]
         [1.27614237 1.86192881]
         [3.19526215 1.58578644]
         [4.         4.        ]]
    """
    # Cumulative distances between consecutive points
    diffs             = np.diff(points, axis=0)
    distances         = np.linalg.norm(diffs, axis=1)
    cumulative_length = np.insert(np.cumsum(distances), 0, 0)

    # Interpolation
    total_length  = cumulative_length[-1]
    new_distances = np.linspace(0, total_length, len(points))

    interp_x = interp1d(cumulative_length, points[:, 0], kind='linear')
    interp_y = interp1d(cumulative_length, points[:, 1], kind='linear')
    
    # Equidistant curve
    new_x = interp_x(new_distances)
    new_y = interp_y(new_distances)
    
    return np.column_stack((new_x, new_y))

def quadratic_bezier(t,P0,P1,P2):
    """
    Evaluates a
    `quadratic Bézier <https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Quadratic_curves>`_ curve
    at parameter ``t`` using three control points.

    Parameters
        t : :class:`float`
            Parameter value between 0.0 and 1.0.
        P0 : :class:`tuple` or :class:`array-like`
            Starting control point (x,y).
        P1 : :class:`tuple` or :class:`array-like`
            Middle control point (x,y).
        P2 : :class:`tuple` or :class:`array-like`
            Ending control point (x,y).
    
    Returns
        :class:`numpy.ndarray` : point (x,y)
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        
        # Evaluate curve at midpoint
        point = chu.quadratic_bezier(0.5, (0, 0), (1, 2), (3, 1))

        print(point)
        
    .. code-block:: python
        :class: out-block

        [1.25 1.25]
    """
    if (P0 is None) or (P1 is None) or (P2 is None): return None
    P0 = np.array(P0); P1 = np.array(P1); P2 = np.array(P2)
    return (1-t)**2 * P0 + 2*(1-t)*t * P1 + t**2 * P2

def get_bezier_curve(P0,P1,P2,n=20):
    """
    Generates a sequence of points along a
    `quadratic Bézier curve <https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Quadratic_curves>`_.

    Parameters
        P0 : :class:`tuple` or :class:`array-like`
            Starting control point (x,y).
        P1 : :class:`tuple` or :class:`array-like`
            Middle control point (x,y).
        P2 : :class:`tuple` or :class:`array-like`
            Ending control point (x,y).
        n : :class:`int`
            Number of points to generate along the curve (default: 20).
    
    Returns
        :class:`numpy.ndarray` : 1D array of points (x,y)
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        
        curve = chu.get_bezier_curve([0, 0], [2, 3], [4, 1], n=10)

        print(curve)
        
    .. code-block:: python
        :class: out-block

        [[0.         0.        ]
         [0.44444444 0.60493827]
         [0.88888889 1.08641975]
         [1.33333333 1.44444444]
         [1.77777778 1.67901235]
         [2.22222222 1.79012346]
         [2.66666667 1.77777778]
         [3.11111111 1.64197531]
         [3.55555556 1.38271605]
         [4.         1.        ]]
    """
    if (P0 is None) or (P1 is None) or (P2 is None): return None
    t_values = np.linspace(0, 1, n)
    return np.array([quadratic_bezier(t, P0, P1, P2) for t in t_values])

def validate_kwargs(keys,allowed_keys,aliases={}):
    """
    Validates that given keyword arguments are within the allowed set of parameters.
    
    Parameters
        keys : :class:`list` or :class:`array-like`
            A list with the name of the key arguments you want to validate.
        allowed_keys : :class:`list` or :class:`array-like`
            A list with the name of the valid arguments.
        aliases : :class:`dict`, optional
            A python dictionary with alternative aliases for the key arguments ().

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu

        user_given_kwargs = {'name': 'Agustina', 'favorite_artist': 'Chappell Roan'}

        allowed_params  = ['name', 'age', 'favorite_color']
        aliases         = {'favorite_color': 'favc'}

        chu.validate_kwargs(user_given_kwargs.keys(),allowed_params,aliases)
    
    .. code-block:: python-console
        :class: out-block

        ---------------------------------------------------------------------------
        Traceback (most recent call last):
            line 8
        KeyError: 'Invalid argument "favorite_artist". Allowed arguments are: name, age, favorite_color / favc.'
    """
    for key in keys:
        if key not in allowed_keys:
            raise KeyError(
            f'Invalid argument "{key}". Allowed arguments are: '
            f'{kwargs_as_string(allowed_keys,aliases)}.'
            )

def kwargs_as_string(keys,aliases={}):
    """
    Formats keyword argument names as a readable string separated by ``,``, with optional aliases.

    Parameters
        keys : :class:`list` or :class:`array-like`
            A list with the name of the key arguments you want to format as string.
        aliases : :class:`dict`, optional
            A python dictionary with alternative aliases for the key arguments ().
    
    Returns
        :class:`str`

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu

        params  = {'name': 'Fabian', 'age': 21, 'favorite_color': 'blue'}
        aliases = {'favorite_color': 'favc'}

        my_kwargs = chu.kwargs_as_string(params.keys(),aliases)
        
        print(f"The arguments of my function are {my_kwargs}.")
    
    .. code-block:: text
        :class: out-block
        
        "The arguments of my function are name, age, favorite_color / favc."
    """
    if not isinstance(keys,list): keys = list(keys)
    for i,key in enumerate(keys):
        if key in aliases.keys(): keys[i] = f'{key} / {aliases[key]}'
    return ', '.join(keys)

def mod_color(color,light=1.0,sat=1.0,alpha=1.0,alpha_bg=(1.0,1.0,1.0)):
    """
    Applies multiple color modifications (lightness, saturation, transparency) to an RGB color.
    
    Parameters
        color : :class:`tuple` or :class:`array-like`
            Triplet with the three RGB values (in arithmetic notation, i.e. 0.0 to 1.0)
        light : :class:`float`
            Lightness factor. 1.0 means no change.
        sat : :class:`float`
            Saturation factor. 1.0 means no change.
        alpha : :class:`float`
            Transparency level (0.0 to 1.0).
        alpha_bg : :class:`tuple` or :class:`array-like`
            Background color. Since transparency isn't actually achieved, a background color is set.
            The color is combined with the background color. Default is white.
    
    Returns
        :class:`tuple` : RGB color
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import matplotlib.colors as mcolors

        # matplotlib.colors.to_rgb also supports HEX codes
        my_color = mcolors.to_rgb("violet")

        # Make it 50% darker, saturate to 150% and set the transparency to 30%
        new_color = chu.mod_color(my_color,light=0.5,sat=1.5,alpha=0.3)

        print("my_color:",my_color)
        print("new_color:",new_color)
    
    .. code-block:: text
        :class: out-block
        
        my_color: (0.9333333333333333, 0.5098039215686274, 0.9333333333333333)
        new_color: (0.8558823529411764, 0.7605882352941176, 0.8558823529411764)
    """
    new_color = color
    # Light
    if light > 1:
        new_color = brighter_color(new_color,factor=light-1)
    elif light < 1:
        new_color = darker_color(new_color,factor=1-light)
    # Saturation
    new_color = saturate_color(new_color,factor=sat)
    # Alpgha
    new_color = alpha_color(new_color,alpha=alpha,bg=alpha_bg)
    return new_color

def brighter_color(color,factor=0.0):
    """
    Increases the brightness of an RGB color by a specified factor.
    
    Parameters
        color : :class:`tuple` or :class:`array-like`
            Triplet with the three RGB values (in arithmetic notation, i.e. 0.0 to 1.0)
        factor : :class:`float`
            Brightness factor. 0.0 means no change.
    
    Returns
        :class:`tuple` : RGB color
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import matplotlib.colors as mcolors

        # matplotlib.colors.to_rgb also supports HEX codes
        my_color = mcolors.to_rgb("violet")

        # Make it 50% brighter
        new_color = chu.brighter_color(my_color,factor=0.5)

        print("my_color:",my_color)
        print("new_color:",new_color)
    
    .. code-block:: text
        :class: out-block
        
        my_color: (0.9333333333333333, 0.5098039215686274, 0.9333333333333333)
        new_color: (0.9666666666666667, 0.7549019607843137, 0.9666666666666667)
    """
    r,g,b = color
    factor = max(0, factor)
    r = min(1, r + (1-r)*factor)
    g = min(1, g + (1-g)*factor)
    b = min(1, b + (1-b)*factor)
    return (r,g,b)

def darker_color(color,factor=0.0):
    """
    Decreases the brightness of an RGB color by a specified factor.
    
    Parameters
        color : :class:`tuple` or :class:`array-like`
            Triplet with the three RGB values (in arithmetic notation, i.e. 0.0 to 1.0)
        factor : :class:`float`
            Darkness factor. 0.0 means no change.
    
    Returns
        :class:`tuple` : RGB color
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import matplotlib.colors as mcolors

        # matplotlib.colors.to_rgb also supports HEX codes
        my_color = mcolors.to_rgb("violet")

        # Make it 50% darker
        new_color = chu.darker_color(my_color,factor=0.5)

        print("my_color:",my_color)
        print("new_color:",new_color)
    
    .. code-block:: text
        :class: out-block
        
        my_color: (0.9333333333333333, 0.5098039215686274, 0.9333333333333333)
        new_color: (0.4666666666666667, 0.2549019607843137, 0.4666666666666667)
    """
    r,g,b = color
    factor = 1 - (max(0,factor))
    r = max(0, r*factor)
    g = max(0, g*factor)
    b = max(0, b*factor)
    return (r,g,b)

def saturate_color(color,factor=1.0):
    """
    Adjusts the saturation level of an RGB color by a specified factor.
    
    Parameters
        color : :class:`tuple` or :class:`array-like`
            Triplet with the three RGB values (in arithmetic notation, i.e. 0.0 to 1.0)
        factor : :class:`float`
            Saturation factor. 1.0 means no change.
    
    Returns
        :class:`tuple` : RGB color
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import matplotlib.colors as mcolors

        # matplotlib.colors.to_rgb also supports HEX codes
        my_color = mcolors.to_rgb("violet")

        # Saturate to 150%
        new_color = chu.saturate_color(my_color,factor=1.5)

        print("my_color:",my_color)
        print("new_color:",new_color)
    
    .. code-block:: text
        :class: out-block
        
        my_color: (0.9333333333333333, 0.5098039215686274, 0.9333333333333333)
        new_color: (1.0, 0.44313725490196076, 0.9999999999999999)
    """
    r,g,b = color
    h,l,s = colorsys.rgb_to_hls(r,g,b)
    s_new = max(0, min(1,s*factor))
    r,g,b = colorsys.hls_to_rgb(h,l,s_new)
    return (r,g,b)

def alpha_color(color, alpha=1.0, bg=(1.0,1.0,1.0)):
    """
    Simulates transparency by blending an RGB color with a background color.
    
    Parameters
        color : :class:`tuple` or :class:`array-like`
            Triplet with the three RGB values (in arithmetic notation, i.e. 0.0 to 1.0)
        alpha : :class:`float`
            Transparency level (0.0 to 1.0).
        bg : :class:`tuple` or :class:`array-like`
            Background color. Since transparency isn't actually achieved, a background color is set.
            The color is combined with the background color. Default is white.
    
    Returns
        :class:`tuple` : RGB color
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu
        import matplotlib.colors as mcolors

        # matplotlib.colors.to_rgb also supports HEX codes
        my_color = mcolors.to_rgb("violet")

        # Set the transparency to 30%
        new_color = chu.alpha_color(my_color,alpha=0.3)

        print("my_color:",my_color)
        print("new_color:",new_color)
    
    .. code-block:: text
        :class: out-block
        
        my_color: (0.9333333333333333, 0.5098039215686274, 0.9333333333333333)
        new_color: (0.98, 0.8529411764705882, 0.98)
    """
    r,g,b = color
    bg_r,bg_g,bg_b = bg
    factor = max(0, min(1, alpha))
    r_result = (r*factor) + (bg_r*(1-factor))
    g_result = (g*factor) + (bg_g*(1-factor))
    b_result = (b*factor) + (bg_b*(1-factor))
    return (r_result,g_result,b_result)

# f-string pre-defined colors
_fstr_colors = {'white':255,'black':232,'light_gray':245,'dark_gray':237,'gold':220,
               'red':196,'blue':21,'green':118,'magenta':165,'mint':87,'orange':202}

def strcol(string,c="white"):
    """
    Applies ANSI color codes to a string (only work in terminal/output cells).
    
    Parameters
        string : :class:`str`
            The string you want to color.
        c : :class:`str` or :class:`int`
            Color to color the string with. This can be a number from the ANSI 8-bit color codes
            (between 0 and 255) or a string of one of the predefined colors: ``"white"``,
            ``"black"``, ``"light_gray"``, ``"dark_gray"``, ``"gold"``,
            ``"red"``, ``"blue"``, ``"green"``, ``"magenta"``, ``"mint"``,
            ``"orange"``.
    
    Returns
        :class:`str`
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.utilities as chu

        print(chu.strcol("Magenta text",c="magenta"))

    .. raw:: html

        <div class="out-block highlight-text notranslate">
        <div class="highlight">
            <pre><span style="color:#d700ff;background:#F5FAF7;font-size:12px">Magenta text</span></pre>
        </div>
        </div>
    """
    if isinstance(c,int):
        if (c >= 0) and (c <= 255): return f'\033[38;5;{c}m{string}\033[0m'
        else: return string
    else:
        if c not in _fstr_colors: return string
        return f'\033[38;5;{_fstr_colors[c]}m{string}\033[0m'

def strcol_palette():
    """
    Displays a visual palette of all available ANSI 256-color codes for terminal text coloring.
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: mock-block

        import cachai.utilities as chu

        chu.strcol_palette()
    """
    for i in range(16):
        for j in range(16):
            print(strcol('■',c=i+j*16) + f' {str(i+j*16):<3.3}  ',end='')
        print()


import pytest
import cachai
import pandas as pd

def test_get_dataset_repo():
	assert cachai.data.get_dataset_repo()=='https://github.com/DD-Beltran-F/cachai-datasets',\
    			'Wrong repository url'

def test_get_dataset_names():
	assert isinstance(cachai.data.get_dataset_names(),list),\
			'Dataset names list not working'

def test_get_dataset_metadata():
	try:
		cachai.data.get_dataset_metadata('lithium')
	except Exception as e:
		pytest.fail(f'get_dataset_metadata went wrong ({e})')

@pytest.mark.parametrize('name', ['lithium','large_correlations','invalid_name'],
					ids=["valid_name:'lithium'","valid_name:'large_correlations'",'invalid_name'])
def test_load_dataset(name):
	try:
		if name == 'invalid_name':
			with pytest.raises(ValueError):
				cachai.data.load_dataset(name)
		else:
			dataset = cachai.data.load_dataset(name)
	except Exception as e:
		pytest.fail(f"load_dataset went wrong when loading '{name}' ({e})")
	
def test_clear_cache():
	# Not-empty cache
	cachai.data.get_dataset_names()
	cachai.data.load_dataset('lithium')
	cachai.data.clear_cache()
	# Empty cache
	cachai.data.clear_cache()


import pytest
import time
import numpy as np
import matplotlib.pyplot as plt
from   cachai.chplot import chord

@pytest.fixture
def sample_corr_matrix():
    return np.array([
        [ 1.0, 0.8, -0.3],
        [ 0.8, 1.0,  0.5],
        [-0.3, 0.5,  1.0]
    ])

@pytest.fixture
def sample_names():
    return ['A', 'B', 'C']

@pytest.fixture
def sample_colors():
    return [(1,0,0), (0,1,0), (0,0,1)]  # Red, Green, Blue

@pytest.fixture
def basic_chord(figure_with_axes, sample_corr_matrix, sample_names, sample_colors):
    fig, ax = figure_with_axes
    return chord(corr_matrix=sample_corr_matrix,
        		 names=sample_names,
        		 colors=sample_colors,
        		 ax=ax)

class TestChordDiagram:
	def test_initialization(self,basic_chord):
		assert basic_chord is not None
		assert len(basic_chord.nodes) == 3

	@pytest.mark.parametrize('invalid_matrix', [
		'not_a_matrix',                   # Not a numpy array
		np.array([1, 2, 3]),              # 1D array
		np.array([[1, 2], [3, 4]]),       # Non-symmetric
		np.array([[1, 2], [2, 'a']]),     # Non-numeric values
		np.array([])                      # Empty array
		])
	def test_corr_matrices(self,invalid_matrix):
		with pytest.raises((ValueError, TypeError)):
		    chord(corr_matrix=invalid_matrix)

	def test_node_count(self,basic_chord):
		assert len(basic_chord.nodes) == basic_chord.corr_matrix.shape[0]

	def test_node_angles(self, basic_chord):
		for node in basic_chord.nodes.values():
		    assert 0 <= node['theta_i'] <= 2*np.pi
		    assert 0 <= node['theta_f'] <= 2*np.pi
		    assert node['theta_i'] < node['theta_f']
		    
	def test_scale_rho_linear(self,basic_chord):
		basic_chord.max_rho = 1.0
		basic_chord.scale   = 'linear'
		assert np.isclose(basic_chord._scale_rho(0.5), 0.5),\
				'rho = 0.5 do not satisfy the linear scale'
		assert basic_chord._scale_rho(1.0) == pytest.approx(1.0),\
				'rho = 1.0 do not satisfy the linear scale'
		assert np.isclose(basic_chord._scale_rho(-0.5), 0.5),\
				'rho = -0.5 do not satisfy the linear scale'

	def test_scale_rho_log(self,basic_chord):
		basic_chord.max_rho = 1.0
		basic_chord.scale = 'log'
		assert basic_chord._scale_rho(0.5) > 0,\
				'rho = 0.5 do not satisfy the log scale'
		assert basic_chord._scale_rho(1.0) == pytest.approx(1.0),\
				'rho = 1.0 do not satisfy the log scale'
		assert basic_chord._scale_rho(0.1) < 0.5,\
				'rho = 0.1 do not satisfy the log scale'

	@pytest.mark.parametrize('threshold,count', [(0.0,3), (0.8,2), (1.0,0)],
							 ids=['threshold=0.0','threshold=0.8','threshold=1.0'])
	def test_generation(self,sample_corr_matrix,sample_names,threshold,count):
		if threshold==1.0:
			with pytest.raises(ValueError):
				chord(corr_matrix=sample_corr_matrix,
		        	  names=sample_names,
		        	  th=threshold)
		else:
			temp_cd = chord(corr_matrix=sample_corr_matrix,
		        		 		names=sample_names,
		        		 		th=threshold)
			assert len(temp_cd.node_patches) == count,\
					'Node patches generation went wrong'
			assert len(temp_cd.node_labels)  == count,\
					'Node labels generation went wrong'
			assert sum(len(chords) for chords in temp_cd.chord_patches) >= count-1,\
					'Chord patches generation went wrong'
			assert sum(len(curves) for curves in temp_cd.bezier_curves) >= count-1,\
					'Bezier curves generation went wrong'

	def test_no_positive_correlations(self):
		cd = chord(
		    corr_matrix=np.array([
		        [1.0, -0.8, -0.5],
		        [-0.8, 1.0, -0.3],
		        [-0.5, -0.3, 1.0]
		    ]),
		    negative_hatch='///'
		)
		plt.close()
		assert cd is not None

	@pytest.mark.parametrize('size,max_time', [(15,2),(20,5),(30,10)],
							 ids=['size=15','size=20','size=30'])
	def test_large_matrix_initialization(self,size,max_time):
		# Matrix
		np.random.seed(42)
		base = np.random.rand(size, size) * 0.8 + 0.1
		large_matrix = (base + base.T) / 2
		np.fill_diagonal(large_matrix, 1.0)

		start_time = time.time()

		temp_cd = chord(corr_matrix=large_matrix,
		    			threshold=0.3,
		    			filter=False)
		initialization_time = time.time() - start_time

		assert temp_cd is not None
		assert len(temp_cd.nodes)  == size
		assert len(temp_cd.names)  == size
		assert len(temp_cd.colors) == size

		total_chords = sum(len(chords) for chords in temp_cd.chord_patches)
		expected_min_chords = size
		assert total_chords >= expected_min_chords, (
		    f'At least {expected_min_chords} connections were expected, '
		    f'but {total_chords} were found'
		)

		assert initialization_time < max_time, 'Initialization took too long'
	
	def test_large_matrix_optimization(self):
		# Matrix
		np.random.seed(42)
		size = 20
		base = np.random.rand(size, size) * 0.8 + 0.1
		large_matrix = (base + base.T) / 2
		np.fill_diagonal(large_matrix, 1.0)

		cd_no_opt = chord(corr_matrix=large_matrix,
		    			  optimize=False)
		cd_opt = chord(corr_matrix=large_matrix,
		    		   optimize=True)

		assert cd_no_opt.order != cd_opt.order,\
				'The optimization did not change the order of the nodes'
		assert set(cd_opt.order) == set(range(size)),\
				'The optimization lost some nodes'

	@pytest.mark.parametrize('node', [0,1,2],
							 ids=['node=0','node=1','node=2'])
	def test_highlighting_nodes(self,sample_corr_matrix,node):
		temp_cd = chord(corr_matrix=sample_corr_matrix,th=0)
		temp_cd.highlight_node(node)

	@pytest.mark.parametrize('node,c', [(0,0),(0,1),(1,0),(1,1),(2,0),(2,1)],
							 ids=['node=0,chord=0','node=0,chord=1',
							      'node=1,chord=0','node=1,chord=1',
							      'node=2,chord=0','node=2,chord=1'])
	def test_highlighting_chords(self,sample_corr_matrix,node,c):
		temp_cd = chord(corr_matrix=sample_corr_matrix,th=0)
		temp_cd.highlight_chord(node,c)



import os
import time
import pandas as pd
import hashlib
import warnings
import json
from   urllib.request import urlopen, Request
from   urllib.error import URLError
from  ._utils import sizeof_fmt, CACHE_DIR, DATASETS_REPO, DATASETS_CATALOG

def _get_cache_path(url):
    """Generate cache path using URL hash"""
    url_hash = hashlib.md5(url.encode()).hexdigest()
    return os.path.join(CACHE_DIR, url_hash)

def _download_with_cache(url, force=False):
    """Download file with persistent cache system"""
    os.makedirs(CACHE_DIR, exist_ok=True)
    cache_path = _get_cache_path(url)
    
    if not force and os.path.exists(cache_path):
        return cache_path

    try:
        req = Request(url, headers={'User-Agent': 'CACHAI'})
        with urlopen(req) as response:
            data = response.read()
            with open(cache_path, 'wb') as f:
                f.write(data)
        return cache_path
    except URLError as e:
        # Fallback to existing cache
        if os.path.exists(cache_path):
            warnings.warn(f"Using cached version due to an error. Details: {str(e)}")
            return cache_path
        raise ConnectionError(f"Error downloading {url}. Details: {str(e)}")

def _get_datasets_catalog(force=False):
    """Obtain the dataset catalog from GitHub"""
    url = DATASETS_REPO + DATASETS_CATALOG
    cached_file = _download_with_cache(url, force)
    
    with open(cached_file, 'r', encoding='utf-8') as f:
        return json.load(f)

def get_dataset_repo():
    """
    Return the URL of the **cachai** datasets repository.

    Returns
        :class:`str`
            GitHub URL of the dataset repository.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.data as chd

        print(chd.get_dataset_repo())

    .. code-block:: text
        :class: out-block

        https://github.com/DD-Beltran-F/cachai-datasets
    """
    return 'https://github.com/DD-Beltran-F/cachai-datasets'

def get_dataset_names():
    """
    Retrieve the list of available dataset names.

    Returns
        :class:`list` of :class:`str`
            Names of the datasets available in the catalog.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.data as chd

        print(chd.get_dataset_names())

    .. code-block:: text
        :class: out-block

        ["lithium", "correlations", "correlations_big"]
    """
    catalog = _get_datasets_catalog(True)
    return list(catalog.keys())

def get_dataset_metadata(name):
    """
    Print the metadata of a specific dataset.

    Parameters
        name : :class:`str`
            Name of the dataset.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.data as chd

        chd.get_dataset_metadata('lithium')

    .. code-block:: text
        :class: out-block

        ══════════════════════════════════════════════════════════════════════════════════════════
        METADATA OF DATASET: LITHIUM
        ──────────────────────────────────────────────────────────────────────────────────────────
        Alias       : lithium
        Filename    : lithium.csv
        Description : Data of lithium abundances and stellar parameters from  M. L. L. Dantas et
                      al. (2025) doi: 10.1051/0004-6361/202453034
        Columns     : CNAME, [Fe/H], A(Li), \\overline{t}_{\\star}, M, T_{eff}, e, Z_{max}, L_z
        ══════════════════════════════════════════════════════════════════════════════════════════
    """
    catalog = _get_datasets_catalog(True)

    if name not in catalog:
        raise ValueError(f"Dataset '{name}' does not exist. "
                         f"The current valid datasets are: {', '.join(get_dataset_names())}.")

    dataset_meta = catalog[name]

    print('═'*50)
    print(f'METADATA OF DATASET: {name.upper()}')
    print('─'*50)
    print(f'Alias       : {name}')
    print(f"Filename    : {dataset_meta.get('filename', 'Not specified')}")
    print(f"Description : {dataset_meta.get('description', 'Not available')}")
    columns = dataset_meta.get('columns', 'Not specified')
    if isinstance(columns,list): columns = ', '.join(columns).replace('$','')
    print(f"Columns     : {columns}")
    print("═"*50 + "\n")

def load_dataset(name="",redownload=False):
    """
    Load a dataset from GitHub with a persistent cache system.

    Parameters
        name : :class:`str`
            Name of the dataset to load.
        redownload : :class:`bool`, optional
            Whether to force re-downloading the dataset, ignoring the cache
            (default: ``False``).

    Returns
        :class:`pandas.DataFrame`
            DataFrame containing the dataset.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.data as chd

        df = chd.load_dataset('lithium')
        print(df.head())

    .. code-block:: text
        :class: out-block

                      CNAME  [Fe/H]  A(Li)  ...         e  Z$_{max}$        $L_z$
        0  00000302-6002570   -0.31   2.01  ...  0.119362   1.003228  2028.535804
        1  00001749-5449565   -0.17   1.98  ...  0.110635   0.670432  1907.144965
        2  00012216-5458205   -0.07   1.51  ...  0.276396   0.996552  1836.529851
        3  00040666-3709129    0.28   0.62  ...  0.112774   0.501939  1676.768299
        4  00042981-4701022   -0.34   1.71  ...  0.257109   1.327526  2239.902280

        [5 rows x 9 columns]
        ...
    """
    catalog = _get_datasets_catalog(redownload)
    
    if name not in catalog:
        raise ValueError(f"Dataset '{name}' does not exist. "
                         f"The current valid datasets are: {', '.join(get_dataset_names())}.")
    
    url = DATASETS_REPO + catalog[name]['filename']
    cached_file = _download_with_cache(url, redownload)
    
    return pd.read_csv(cached_file)

def clear_cache(max_age_days=0):
    """
    Delete old cached files from **cachai**'s cache directory.

    Parameters
        max_age_days : :class:`int`, optional
            Maximum file age in days. Files older than this will be deleted.
            If set to ``0`` (default), all files are removed.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: in-block

        import cachai.data as chd

        # Delete all cached files
        chd.clear_cache()

    .. code-block:: text
        :class: out-block

        3 file(s) deleted by 2025-09-29 (14:43:47) from cachai's cache folder.
        Space freed: 153.1 KB (100.0%).
    
    You can also choose the maximum file age in days, files older than ``max_age_days`` will be
    deleted:

    .. code-block:: python
        :class: mock-block

        import cachai.data as chd

        # Delete files older than 30 days
        clear_cache(max_age_days=30)
    """
    now         = time.time()
    counts      = 0
    total       = 0
    freed_space = 0
    
    total_space = sum(
        os.path.getsize(os.path.join(CACHE_DIR, f)) 
        for f in os.listdir(CACHE_DIR) 
        if os.path.isfile(os.path.join(CACHE_DIR, f)))

    if total_space == 0:
        print("cachai's cache folder is already empty.")
        return

    if os.path.exists(CACHE_DIR):
        total = len(os.listdir(CACHE_DIR))
        for filename in os.listdir(CACHE_DIR):
            filepath = os.path.join(CACHE_DIR, filename)
            if os.stat(filepath).st_mtime < now - max_age_days * 86400:
                file_size = os.path.getsize(filepath)
                try:
                    os.remove(filepath)
                    counts += 1
                    freed_space += file_size
                except Exception as e:
                    warnings.warn(f'Could not delete {filepath}. Details: {str(e)}')
    
    now_str = time.strftime("%Y-%m-%d (%H:%M:%S)", time.localtime(now))
    print(f"{counts} file(s) deleted by {now_str} from cachai's cache folder.\n"
          f'Space freed: {sizeof_fmt(freed_space)} ({freed_space/total_space*100:.1f}%).')


import sys
import matplotlib.pyplot as plt
from   pathlib import Path
from   importlib.util import find_spec

def get_available_tests():
    """
    Retrieve the list of available test modules.

    Returns
        :class:`list` of :class:`str`
            Names of the available tests (retrieved from the test files without the
            ``test_`` prefix).

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    If you are running tests in the terminal, you can run ``cachai-test`` using the ``-l`` or
    ``--list`` option:
    
    .. code-block:: console
        :class: in-block

        $ cachai-test -l
    
    .. code-block:: console
        :class: out-block

        Available tests:
            utilities
            data
            dependencies
            gadgets
            charts
    
    If you are working in a Python/Jupyter file:

    .. code-block:: python
        :class: in-block

        import cachai as ch

        tests = ch.get_available_tests()
        print("Available tests:", tests)

    .. code-block:: text
        :class: out-block

        Available tests: ['utilities', 'data', 'dependencies', 'gadgets', 'charts']
    
    """
    test_dir = Path(__file__).parent
    return [f.stem[5:] for f in test_dir.glob('test_*.py')]

def run_tests(*test_args,show_details=False):
    """
    Run **cachai**'s test suite using pytest.

    Parameters
        test_args : :class:`str`
            Names of the tests to run (e.g., ``'charts'``, ``'utilities'``).
            If not provided, all available tests will be executed.
        show_details : :class:`bool`, optional
            Whether to include verbose output (``-v``) in pytest execution
            (default: ``True``).
    
    Returns
        :class:`int`
            Exit code returned by pytest.
    
    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    There are two ways to run the cachai tests. It's recommended to do so from the terminal,
    as follows:

    .. code-block:: console
        :class: in-block

        $ cachai-test
    
    .. code-block:: console
        :class: out-block

        ================================ test session starts =================================
        platform linux -- Python 3.12.3, pytest-8.4.1, pluggy-1.6.0 -- /path_to_interpreter
        rootdir: /home/user
        plugins: anyio-4.10.0
        collected 57 items

        path_to_packages/cachai/tests/test_charts.py ...........................        [ 47%]
        path_to_packages/cachai/tests/test_data.py .......                              [ 59%]
        path_to_packages/cachai/tests/test_dependencies.py .....                        [ 68%]
        path_to_packages/cachai/tests/test_gadgets.py .....                             [ 77%]
        path_to_packages/cachai/tests/test_utilities.py .............                   [100%]

        ================================ 57 passed in 11.89s =================================
    
    You can also run a specific test from the list of available tests, and select the verbose
    option for more details.

    .. code-block:: console
        :class: in-block

        $ cachai-test utilities
    
    .. code-block:: console
        :class: out-block

        ================================= test session starts ==================================
        platform linux -- Python 3.12.3, pytest-8.4.1, pluggy-1.6.0 -- /path_to_interpreter
        cachedir: .pytest_cache
        rootdir: /home/user
        plugins: anyio-4.10.0
        collected 13 items                                                                                                                                                                   

        path_to_packages/cachai/tests/test_utilities.py::test_validate_kwargs PASSED      [  7%]
        path_to_packages/cachai/tests/test_utilities.py::test_save_func PASSED            [ 15%]
        path_to_packages/cachai/tests/test_utilities.py::test_rgb_hsl_conversions PASSED  [ 23%]
        path_to_packages/cachai/tests/test_utilities.py::test_mod_color PASSED            [ 30%]
        path_to_packages/cachai/tests/test_utilities.py::test_angspace PASSED             [ 38%]
        path_to_packages/cachai/tests/test_utilities.py::test_angdist[pi->pi] PASSED      [ 46%]
        path_to_packages/cachai/tests/test_utilities.py::test_angdist[3pi/2->pi/2] PASSED [ 53%]
        path_to_packages/cachai/tests/test_utilities.py::test_angdist[2pi->0] PASSED      [ 61%]
        path_to_packages/cachai/tests/test_utilities.py::test_quadratic_bezier PASSED     [ 69%]
        path_to_packages/cachai/tests/test_utilities.py::test_get_bezier_curve PASSED     [ 76%]
        path_to_packages/cachai/tests/test_utilities.py::test_equidistant PASSED          [ 84%]
        path_to_packages/cachai/tests/test_utilities.py::test_map_from_curve PASSED       [ 92%]
        path_to_packages/cachai/tests/test_utilities.py::test_colormapped_patch PASSED    [100%]

        ================================== 13 passed in 1.11s ==================================

    The other option is to run the tests in a Python/Jupyter file.

    .. code-block:: python
        :class: mock-block

        import cachai as ch

        # Run all tests verbose
        ch.run_tests(show_details=True)

    .. code-block:: python
        :class: mock-block

        import cachai as ch

        # Run specific tests (e.g. charts and data)
        ch.run_tests('charts','data')
    
    """
    if find_spec('pytest') is None:
        print('\nError: To run the tests you need to have pytest installed.', file=sys.stderr)
        print('You can install it with:', file=sys.stderr)
        print('\n    pip install pytest\n', file=sys.stderr)
        return None
    import pytest
    
    test_dir = Path(__file__).parent
    pytest_args = ['-p', 'no:asdf_schema_tester']
    if show_details:
        pytest_args.append('-v')

    if test_args:
        # Validation
        missing = []
        valid_tests = []
        for arg in test_args:
            test_file = test_dir / f'test_{arg}.py'
            if test_file.exists():
                valid_tests.append(str(test_file))
            else:
                missing.append(arg)
        
        if missing:
            raise ValueError(f'Tests not found: {", ".join(missing)}. '
                           f'The available tests are: {get_available_tests()}')
        
        pytest_args.extend(valid_tests)
    else:
        pytest_args.append(str(test_dir))

    plt.close('all')
    return pytest.main(pytest_args)


import os
import pytest
import numpy as np
import matplotlib.pyplot as plt
from   matplotlib.patches import Circle
import cachai.utilities as chu

@pytest.fixture
def sample_color():
    return (0.2, 0.4, 0.6)

@pytest.fixture
def sample_curve():
    return np.array([[0, 0], [1, 1], [2, 0]])

def test_validate_kwargs():
	params = {'param1': 1, 'param2': 2}
	# Test valid kwargs
	chu.validate_kwargs(['param1'],params)
	# Test invalid kwargs
	with pytest.raises(KeyError):
		chu.validate_kwargs(['invalid_param'],params)

def test_chsave(tmp_path):
    # Simple figure
    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 1])
    # Testing paths
    test_dir     = tmp_path / 'tmp_images'
    test_pdf_dir = os.path.join(test_dir,'pdf')
    test_name    = 'tmp_img'
    try: chu.chsave(fig_name=test_name,path=test_dir,pdf=True)
    except: pass
    assert os.path.exists(test_dir),\
    			'Main directory was not created'
    assert os.path.exists(test_pdf_dir),\
    			'PDF subdirectory was not created'
    assert os.path.exists(os.path.join(test_dir,f'{test_name}.png')),\
    			'PNG file was not created'
    assert os.path.exists(os.path.join(test_pdf_dir,f'{test_name}.pdf')),\
    			'PDF file was not created'
    plt.close(fig)

def test_mod_color(sample_color):
    try:
        brighter  = chu.mod_color(sample_color, light=1.5)
        darker    = chu.mod_color(sample_color, light=0.5)
        saturated = chu.mod_color(sample_color, sat=2)
        trans     = chu.mod_color(sample_color, alpha=0.5)
    except Exception as e:
        pytest.fail(f'Color was not modified correctly ({e})')
    assert np.mean(brighter) > np.mean(sample_color) > np.mean(darker),\
            'Color light was not modified correctly'
    assert np.all(saturated != sample_color),\
            'Color saturation was not modified correctly'
    assert np.all(trans != sample_color),\
            'Color transparency was not modified correctly'

def test_angspace():
    angles = chu._angspace(0, np.pi/2, n=100)
    assert len(angles) == 25
    assert np.isclose(angles[-1], np.pi/2)

@pytest.mark.parametrize('inp,out', [
        (np.pi, pytest.approx(np.pi)),
        (3*np.pi/2, pytest.approx(np.pi/2)),
        (2*np.pi, pytest.approx(0)),
    ], ids=['pi->pi','3pi/2->pi/2','2pi->0'])
def test_angdist(inp,out):
    assert chu.angdist(0, inp) == out

def test_quadratic_bezier(sample_curve):
    P0, P1, P2 = sample_curve
    assert np.allclose(chu.quadratic_bezier(0, P0, P1, P2), P0)
    assert np.allclose(chu.quadratic_bezier(1, P0, P1, P2), P2)

def test_get_bezier_curve(sample_curve):
    P0, P1, P2 = sample_curve
    curve      = chu.get_bezier_curve(P0, P1, P2, n=10)
    assert curve.shape == (10, 2)
    assert np.allclose(curve[0], P0)
    assert np.allclose(curve[-1], P2)

def test_equidistant(sample_curve):
    eq_curve  = chu.equidistant(sample_curve)
    diffs     = np.diff(eq_curve, axis=0)
    distances = np.linalg.norm(diffs, axis=1)
    assert np.allclose(distances, distances[0], rtol=1e-2)

def test_map_from_curve(sample_curve):
    P0, P1, P2 = sample_curve
    curve      = chu.get_bezier_curve(P0, P1, P2, n=10)
    map_mat    = chu.map_from_curve(curve, resolution=50)
    assert map_mat.shape == (50, 50)
    assert np.min(map_mat) >= -1
    assert np.max(map_mat) <= 1

def test_colormapped_patch(sample_curve):
    fig, ax = plt.subplots()
    patch   = Circle((0.5, 0.5), 0.4)
    ax.add_patch(patch)
    
    P0, P1, P2 = sample_curve
    curve      = chu.get_bezier_curve(P0, P1, P2, n=10)
    map_mat    = chu.map_from_curve(curve, resolution=50)
    img        = chu.colormapped_patch(patch, map_mat, ax=ax)
    
    assert img.get_array().shape == (50, 50)
    plt.close(fig)

if __name__ == '__main__':
    pytest.main(['-v', __file__])


import pytest
import importlib
from   packaging import requirements

# List of dependencies and requirements
DEPENDENCIES = [
    'matplotlib>=3.10.3',
    'numpy>=1.2.6',
    'pandas>=2.3.0',
    'seaborn>=0.13.2',
    'scipy>=1.15.3',
]

@pytest.mark.parametrize('dep', DEPENDENCIES)
def test_dependency(dep):
    req = requirements.Requirement(dep)
    try:
        mod = importlib.import_module(req.name)
        if req.specifier:
            assert req.specifier.contains(mod.__version__), \
                f'Your {req.name} version is {mod.__version__}, the requirement is {req.specifier}'
    except ImportError:
        pytest.fail(f'{req.name} is not installed')


# Basic imports
import numpy as np
# Matplotlib imports
from   matplotlib.text import Text

class PolarText(Text):
    """
    Initialize a :class:`matplotlib.text.Text` instance using polar coordinates.

    Attributes
        radius : :class:`float`
            Radius coordinate from the center.
        theta : :class:`float`
            Angle in degrees (0° at positive x-axis, increasing counterclockwise).
        text : :class:`str`
            Text to diplay.
        center : :class:`tuple`, optional
            Center coordinates (x,y) (default: (0,0)).
        pad : :class:`float`, optional
            Padding value (positive for outward displacement, negative for inward).

    Other Attributes
        ``**kwargs``
            Keyword arguments of :class:`matplotlib.text.Text`.

    Examples
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. code-block:: python
        :class: mock-block

        import matplotlib.pyplot as plt
        import cachai.gadgets as chg

        fig, ax = plt.subplots()
        ax.plot([0,0],[1,1])
        polar = chg.PolarText((0,0),1,45,text='Im polar!',pad=0.5,ha='center', va='center')
        ax.add_artist(polar)
        plt.show()
    
    Methods
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
    """
    def __init__(self, radius, angle, text, center=(0,0), pad=0.0, **kwargs):
        """:meta private:"""
        self._radius = radius
        self._angle  = np.deg2rad(angle)
        self.center  = np.array(center)
        self._pad    = pad
        
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        
        super().__init__(x, y, text, **kwargs)

    def _polar_to_cartesian(self, radius, angle):
        """:meta private:
        Tranform polar (radius, angle) to catesian (x, y).
        
            radius : :class:`float`
                Radius coordinate.
            angle : :class:`float`
                Angle coordinate in radians.
        """
        dx = radius * np.cos(angle)
        dy = radius * np.sin(angle)
        return self.center + np.array([dx, dy])

    def set_polar_position(self, radius=None, angle=None, center=None):
        """
        Updates multiple polar coordinates simultaneously.

        Parameters
            radius : :class:`float`
                New radial distance from center.
            angle : :class:`float`
                New angular position in degrees.
            center : :class:`tuple` or :class:`array-like`
                New center coordinates for the polar system.
        """
        if radius is not None:
            self._radius = radius
        if angle is not None:
            self._angle = np.deg2rad(angle)
        if center is not None:
            self.center = np.array(center)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))
    
    def set_pad(self, pad):
        """
        Adjusts the radial padding distance between the text and its base polar position.
        
        Parameters
            pad : :class:`float`
                New padding value.
        """
        self._pad = pad
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))
    
    def set_radius(self, radius):
        """
        Updates the radial coordinate of the text relative to the center point.
        
        Parameters
            radius : :class:`float`
                New radius coordinate.
        """
        self._radius = radius
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))

    def set_angle(self, angle):
        """
        Changes the angular position of the text around the center.
        
        Parameters
            angle : :class:`float`
                New angle coordinate in degrees.
        """
        self._angle = np.deg2rad(angle)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))

    def set_center(self, center):
        """
        Moves the entire polar coordinate system to a new center point.

        Parameters
            center : :class:`tuple` or :class:`array-like`
                New center coordinates (x,y).
        """
        self.center = np.array(center)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))


import os
from   pathlib import Path

CACHE_DIR        = os.path.join(Path.home(), '.cachai_cache')
DATASETS_REPO    = 'https://raw.githubusercontent.com/DD-Beltran-F/cachai-datasets/main/'
DATASETS_CATALOG = 'datasets_catalog.json'

def sizeof_fmt(num,suffix='B'):
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"